"""PRM Implementations for Local HuggingFace Backends."""

import torch
from transformers.tokenization_utils_base import BatchEncoding

from mellea.backends.huggingface import HFProcessRewardModel


class HFGenerativePRM(HFProcessRewardModel):
    """A Generative PRM that works with a huggingface backend."""

    def __init__(
        self,
        model_name_or_path: str = "ibm-granite/granite-3.3-8b-lora-math-prm",
        score_token: str = "Y",
        device: str | None = None,
        generation_prompt: str = "Is this response correct so far (Y/N)?",
        step_separator: str = "\n\n",
    ):
        """Initialize a Generative PRM that works with a huggingface backend. Currently supports and tested with IBM Process Reward Models.

        Args:
            model_name_or_path (str): A local path to PRM or a huggingface PRM
            score_token (str): token who's logits correspond to the PRM score. Usually is a correctness indicator (for generative PRMs)
            device (str): pointer to device
            generation_prompt (str): Optional prompt to be added before generation
            step_separator (str): string on which to separate the content into steps
        """
        super().__init__(model_name_or_path, score_token, device)
        self.generation_prompt = (
            generation_prompt if generation_prompt is not None else ""
        )
        self.step_separator = step_separator
        self.softmax = torch.nn.Softmax(dim=-1)

    def score(self, query: str, response: str) -> tuple[list[float], list[list[float]]]:
        """Returns a final and per-step score for a given input query and response.

        Args:
            query (str): User query
            response (str): Assistant Response to score
        """
        list_of_steps = self.stepify(response, self.step_separator)
        # get tokenized batch
        batches = self.prepare_inputs(query, list_of_steps)
        all_rewards = []
        all_rewards_per_step = []

        # find the chat turn where assistant message starts to find the correct placement of the score token
        # add empty system prompt to prevent model from adding its own system prompt
        chat_template_to_turn = self.tokenizer.apply_chat_template(
            [
                {"role": "system", "content": ""},
                {"role": "assistant", "content": self._score_token},
            ],
            tokenize=False,
            add_generation_prompt=False,
        )
        # removing the system prompt by finding the assistant turn, which usually starts like <|..|>assistant<|..>
        asst_text = chat_template_to_turn[chat_template_to_turn.find(">assistant<") :][
            1:
        ]
        asst_toks = self.tokenizer(
            asst_text, add_special_tokens=False, return_tensors="pt"
        )["input_ids"][0]
        asst_toks_before_correct_token = asst_toks[
            : torch.where(asst_toks == self._score_token_id)[
                0
            ].item()  # type: ignore
        ].tolist()  # type : ignore

        # move each item of the batch to the device
        for i in batches:
            batches[i] = batches[i].to(self.model.device)

        with torch.no_grad():
            model_outputs = self.model(**batches)
            logits = model_outputs.logits  # (bsz, seq_len, vocab_size)

        for batch_idx in range(logits.shape[0]):
            per_input_rewards = []
            # for each element in the batch (i.e., each input)
            # we need to get logits for all tokens where the token is self._score_token (in assistant turn)
            # find batch index for **assistant** turn is self._score_token, not just the self._score_token_id
            correct_token_indices = torch.where(
                batches["input_ids"][batch_idx] == self._score_token_id
            )[0].tolist()
            prm_indices = []
            for t_idx in correct_token_indices:
                if (
                    batches["input_ids"][batch_idx][
                        t_idx - len(asst_toks_before_correct_token) : t_idx
                    ].tolist()
                    == asst_toks_before_correct_token
                ):
                    prm_indices.append(
                        t_idx - 1
                    )  # the logits for token i predict the token i+1: so, we need to look at the **previous** token logits

            assert len(prm_indices) > 0
            #  convert logits to probabilities and get the probability of the correct token id as reward
            for prm_idx in prm_indices:
                per_input_rewards.append(
                    self.softmax(logits[batch_idx, prm_idx, :])[
                        self._score_token_id
                    ].item()
                )

            # aggregate. return final rewards
            all_rewards_per_step.append(per_input_rewards)
            sum = 0
            for reward in per_input_rewards:
                sum += reward
            per_input_reward = sum / len(per_input_rewards)
            all_rewards.append(per_input_reward)

        return all_rewards, all_rewards_per_step

    def prepare_inputs(self, user_content: str, steps: list[str]) -> BatchEncoding:
        """Prepare the inputs for inference with the model.

        Args:
            user_content (str): the user query
            steps (List(str)): assistant response, broken down into steps
        """
        msgs = []
        for s_idx, step in enumerate(steps):
            # apply chat template as expected by the reward model
            # rewards are calculated from the logit of self._score_token as produced by the assistant
            if s_idx == 0:
                msgs.append(
                    {
                        "role": "user",
                        "content": user_content
                        + " "
                        + step
                        + " "
                        + self.generation_prompt,
                    }
                )
            else:
                # first add last assistant turn
                msgs.append({"role": "assistant", "content": self._score_token})
                msgs.append(
                    {"role": "user", "content": step + " " + self.generation_prompt}
                )

        # append last assistant turn
        msgs.append({"role": "assistant", "content": self._score_token})
        input_message = self.tokenizer.apply_chat_template(
            msgs, add_generation_prompt=False, tokenize=False
        )
        return self.tokenizer(
            [input_message], return_tensors="pt", padding=True, truncation=True
        )


class HFRegressionPRM(HFProcessRewardModel):
    """A Regression PRM that works with a huggingface backend."""

    def __init__(
        self,
        model_name_or_path: str,
        score_token: str = "<end_of_step>",
        device: str | None = None,
        step_separator: str = "\n\n",
    ):
        """Initialize a Regression PRM that works with a huggingface backend. Currently supports and tested with IBM Process Reward Models.

        Args:
            model_name_or_path (str): A local path to PRM or a huggingface PRM
            score_token (str): token who's logits correspond to the PRM score. Usually is a step demarker (for non-generative PRMs)
            device (str): pointer to the device on which to run the model
            step_separator (str): string on which to separate the input content into steps
        """
        super().__init__(model_name_or_path, score_token, device)

        # initialize PRM head
        self.prm_head = torch.nn.Linear(
            self.model.config.hidden_size, 2, bias=False, dtype=self.model.dtype
        ).to(self.model.device)

        state = torch.load(model_name_or_path + "/added_params.bin")
        # need to do this-- we save model dict as `prm_head.weight` during training
        new_state_dict = {}
        for k, v in state.items():
            new_k = k.replace("prm_head.", "")
            new_state_dict[new_k] = v

        self.prm_head.load_state_dict(new_state_dict)
        self.prm_head.eval()

        self.step_separator = step_separator
        self.softmax = torch.nn.Softmax(dim=-1)

    def score(self, query: str, response: str) -> tuple[list[float], list[list[float]]]:
        """Returns a final and per-step score for a given input query and response.

        Args:
            query (str): User query
            response (str): Assistant Response to score
        """
        list_of_steps = self.stepify(response, self.step_separator)
        # tokenizes the batch and concatenates the list of steps into a single step-separated response
        batch = self.prepare_inputs(query, list_of_steps)
        # move each item of the batch to the device
        for i in batch:
            batch[i] = batch[i].to(self.model.device)

        with torch.no_grad():
            model_outputs = self.model(**batch, output_hidden_states=True)
            # all logits
            all_prm_logits = self.prm_head(model_outputs["hidden_states"][-1]).squeeze(
                -1
            )

        # get logits for each end of step i.e. logits for step_eos positions in the input
        prm_probs = []
        rewards = []
        for idx in range(all_prm_logits.shape[0]):
            prm_indices = torch.where(batch["input_ids"][idx] == self._score_token_id)[
                0
            ]
            assert prm_indices.shape[0] > 0
            # head produces two logits, the second one is the logit for the correct answer
            # convert logits to probabilities using softmax
            # return list of floats instead of list of tensors
            prm_probs_per_sample = [
                t.item() for t in self.softmax(all_prm_logits[idx][prm_indices])[:, 1]
            ]
            prm_probs.append(prm_probs_per_sample)

            reward = sum(prm_probs_per_sample) / len(prm_probs_per_sample)
            rewards.append(reward)

        return rewards, prm_probs

    def prepare_inputs(self, user_content: str, steps: list[str]) -> BatchEncoding:
        """Prepare the inputs for inference with the model.

        Args:
            user_content (str): the user query
            steps (List(str)): assistant response, broken down into steps
        """
        text_with_steps_marked = ""

        for step in steps:
            text_with_steps_marked += f"{step} {self._score_token}"

        message = [
            {"role": "user", "content": user_content},
            {"role": "assistant", "content": text_with_steps_marked},
        ]
        input_message = self.tokenizer.apply_chat_template(message, tokenize=False)

        return self.tokenizer(
            [input_message], return_tensors="pt", padding=True, truncation=True
        )



"""Huggingface implementations for IBM's "starter pack" of Activated LoRAs."""

import asyncio
import functools
from copy import deepcopy

import torch
from transformers.generation.utils import GenerateDecoderOnlyOutput

from mellea.backends.huggingface import HFAlora, HFAloraCacheInfo, LocalHFBackend
from mellea.backends.types import ModelOption
from mellea.helpers.async_helpers import send_to_queue
from mellea.helpers.fancy_logger import FancyLogger
from mellea.stdlib.base import GenerateType, ModelOutputThunk


class HFConstraintAlora(HFAlora):
    """The Requirement Checking ALora for Granite checks if the specified requirement was satisfied by the most recent model generation. Only one requirement is checked at a time.

    Currently supports [Granite 3.2 8B](https://huggingface.co/ibm-granite/granite-3.2-8b-alora-requirement-check) and [Granite 3.3 8B](https://huggingface.co/ibm-granite/granite-3.3-8b-alora-requirement-check) by default.
    """

    def __init__(
        self,
        name: str,
        path_or_model_id: str,
        generation_prompt: str,
        backend: LocalHFBackend,
        *,
        constraint_prompt: str | None = None,
        include_constraint_in_alora_offset: bool = False,
    ):
        """Initialize after checking that the backend is correct.

        Args:
            name: name of the alora.
            path_or_model_id: huggingface path or model id.
            generation_prompt: the prompt required to activate the aLoRa.
            backend: a LocalHFBackend that this alora is attached to.
            constraint_prompt: a template that the constraint can be interpolated into; can only have a single `{}` slot.
            include_constraint_in_alora_offset: whether to include the constraint prompt in the alora offset.
        """
        super().__init__(name, path_or_model_id, generation_prompt, backend)

        # Maintain default behavior.
        if constraint_prompt is None:
            constraint_prompt = "\nRequirement: {}<|end_of_text|>\n"

        self._constraint_prompt = constraint_prompt
        self._include_constraint_in_alora_offset = include_constraint_in_alora_offset

        # We do a lot of logging for ALoras because this is an experimental feature. Maybe we should tag these log messages?
        self._logger = FancyLogger.get_logger()

    def generate_using_strings(
        self,
        input: str,
        response: str,
        constraint: str,
        force_yn: bool = True,
        stream: bool = False,
    ) -> ModelOutputThunk:
        """Generates a constraint response from the ALora. Must be run in a running event loop."""
        assert self._backend.alora_model is not None
        # Go ahead and do runtime type-checking because passing CBlocks into this function is a common error.
        assert type(input) is str
        assert type(response) is str
        assert type(constraint) is str
        self._backend.alora_model.set_adapter(self.name)
        cache_hit = self._backend.cache_get(response)

        if stream:
            self._logger.warning(
                "`HFConstraintAlora` cannot stream output; defaulting to non-streaming approach."
            )

        generate_kwargs = {}
        if cache_hit:
            self._logger.debug(
                f"using cache for alora {self.__class__} and response '{response}'"
            )
            generate_kwargs["past_key_values"] = deepcopy(cache_hit.kv_cache)
            input_combined = self._generate_using_cache(cache_hit, constraint, force_yn)

        else:
            self._logger.debug(
                f"not using cache for alora {self.__class__} and response '{response}'"
            )
            input_combined = self._generate_not_using_cache(
                input, response, constraint, force_yn
            )

        if not self._include_constraint_in_alora_offset:
            alora_offsets = [self._generation_prompt_tokens["input_ids"].shape[1] - 1]
        else:
            # Get the constraint tokens separately so that we can calculate the alora offsets.
            constraint_tokens = self._backend._tokenizer(
                self._constraint_prompt.format(constraint), return_tensors="pt"
            ).to(self._backend._device)

            alora_offsets = [
                constraint_tokens["input_ids"].shape[1]
                + self._generation_prompt_tokens["input_ids"].shape[1]
                - 2
            ]

        chat_response = asyncio.to_thread(
            self._backend.alora_model.generate,
            input_combined["input_ids"].to(self._backend._device),
            attention_mask=input_combined["attention_mask"].to(self._backend._device),
            max_new_tokens=1,
            return_dict_in_generate=True,
            alora_offsets=alora_offsets,
            output_scores=True,
            **generate_kwargs,
        )

        output = ModelOutputThunk(None)
        output._meta["alora_name"] = self.name

        output._process = functools.partial(
            processing,
            backend=self._backend,
            force_yn=force_yn,
            gen_prompt=self._generation_prompt,
        )
        output._post_process = functools.partial(post_processing, backend=self._backend)

        try:
            # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
            # We can also support synchronous calls by adding a flag and changing this ._generate function.

            # This function should always be called from a running event loop so we don't have to worry about
            # scheduling the task to a specific event loop here.
            output._generate = asyncio.create_task(
                send_to_queue(chat_response, output._async_queue)  # type: ignore
            )
            output._generate_type = GenerateType.ASYNC
        except RuntimeError as e:
            # Most likely cause is running this function without an event loop present.
            raise e

        return output

    def _generate_using_cache(
        self, cache_hit: HFAloraCacheInfo, constraint: str, force_yn: bool
    ) -> dict:
        """Returns the input object used for generation."""
        # Must tokenize the constraint here since the requirement isn't known at initialization.
        constraint_tokens = self._backend._tokenizer(
            self._constraint_prompt.format(constraint), return_tensors="pt"
        ).to(self._backend._device)

        input_combined = {
            "input_ids": torch.cat(
                [
                    cache_hit.merged_token_ids.unsqueeze(0),
                    constraint_tokens["input_ids"],
                    self._generation_prompt_tokens["input_ids"],
                ],
                dim=1,
            ),
            "attention_mask": torch.cat(
                [
                    cache_hit.merged_attention.unsqueeze(0),
                    constraint_tokens["attention_mask"],
                    self._generation_prompt_tokens["attention_mask"],
                ],
                dim=1,
            ),
        }

        self._logger.debug(
            f"Prompt for cached aLoRA({self.name}):\n {self._backend._tokenizer.decode(input_combined['input_ids'][0])}"
        )

        return input_combined

    def _generate_not_using_cache(
        self, input: str, response: str, constraint: str, force_yn: bool
    ) -> dict:
        """Returns the input object used for generation."""
        # Params aren't needed when just getting the backend args.
        backend_model_opts = self._backend._simplify_and_merge(None)
        sys_prompt = backend_model_opts.get(ModelOption.SYSTEM_PROMPT, None)

        chat = [
            *([{"role": "system", "content": sys_prompt}] if sys_prompt else []),
            {"role": "user", "content": input},
            {"role": "assistant", "content": response},
        ]

        templatized = self._backend._tokenizer.apply_chat_template(chat, tokenize=False)
        assert type(templatized) is str

        # Must tokenize the constraint here since the requirement isn't known at initialization.
        templatized = templatized + self._constraint_prompt.format(constraint)

        tokenized = self._backend._tokenizer(templatized, return_tensors="pt").to(
            self._backend._device
        )

        input_combined = {
            "input_ids": torch.cat(
                [tokenized["input_ids"], self._generation_prompt_tokens["input_ids"]],
                dim=1,
            ),
            "attention_mask": torch.cat(
                [
                    tokenized["attention_mask"],
                    self._generation_prompt_tokens["attention_mask"],
                ],
                dim=1,
            ),
        }

        self._logger.debug(
            f"Prompt for non-cached aLoRA({self.name}):\n{self._backend._tokenizer.decode(input_combined['input_ids'][0])}"
        )

        return input_combined


async def processing(
    mot: ModelOutputThunk,
    chunk: GenerateDecoderOnlyOutput,
    backend: LocalHFBackend,
    force_yn: bool,
    gen_prompt: str,
):
    """Called to process the incoming chunks."""
    if mot._underlying_value is None:
        mot._underlying_value = ""

    # Don't support async for HFConstraintAlora. Means we can process the output here.
    assert isinstance(chunk, GenerateDecoderOnlyOutput)

    if force_yn:
        last_logits = chunk.scores[-1].squeeze(0)  # type: ignore
        token_Y = backend._tokenizer("Y", add_special_tokens=False)["input_ids"][0]  # type: ignore
        token_N = backend._tokenizer("N", add_special_tokens=False)["input_ids"][0]  # type: ignore
        logit_Y = last_logits[token_Y].item()
        logit_N = last_logits[token_N].item()
        mot._underlying_value = "Y" if logit_Y > logit_N else "N"
    else:
        output_text = backend._tokenizer.decode(chunk.sequences[0])
        constraint_satisfied = output_text.split(gen_prompt)[-1]
        mot._underlying_value = constraint_satisfied[
            0
        ]  # Grab the first char of the str.


async def post_processing(mot: ModelOutputThunk, backend: LocalHFBackend):
    """Called after all data has been received."""
    backend.formatter.parse(mot._action, mot)  # type: ignore


def add_granite_aloras(backend: LocalHFBackend):
    """Adds the IBM Granite "starter pack" ALoras to a backend."""
    if backend._hf_model_id == "ibm-granite/granite-3.2-8b-instruct":
        backend.add_alora(
            HFConstraintAlora(
                name="constraint",
                path_or_model_id="ibm-granite/granite-3.2-8b-alora-requirement-check",
                generation_prompt="<|start_of_role|>check_requirement<|end_of_role|>",
                backend=backend,
                constraint_prompt="\nRequirement: {}<|end_of_text|>\n",
                include_constraint_in_alora_offset=False,
            )
        )
    elif backend._hf_model_id == "ibm-granite/granite-3.3-8b-instruct":
        backend.add_alora(
            HFConstraintAlora(
                name="constraint",
                path_or_model_id="ibm-granite/granite-3.3-8b-alora-requirement-check",
                generation_prompt="<|start_of_role|>check_requirement<|end_of_role|>",
                backend=backend,
                constraint_prompt="\n<|start_of_role|>requirement<|end_of_role|>{}<|end_of_text|>\n",
                include_constraint_in_alora_offset=True,
            )
        )
    else:
        raise ValueError(
            f"cannot add_granite_aloras to unknown huggingface model_id / backend: {backend._hf_model_id}"
        )



"""Best of N Sampling Strategy."""

from copy import deepcopy

import tqdm

from mellea.backends import Backend, BaseModelSubclass
from mellea.helpers.async_helpers import wait_for_all_mots
from mellea.helpers.fancy_logger import FancyLogger
from mellea.stdlib import funcs as mfuncs
from mellea.stdlib.base import CBlock, ChatContext, Component, Context, ModelOutputThunk
from mellea.stdlib.instruction import Instruction
from mellea.stdlib.requirement import Requirement, ScorerRequirement, ValidationResult
from mellea.stdlib.sampling import BaseSamplingStrategy, SamplingResult


class BestofNSamplingStrategy(BaseSamplingStrategy):
    """Sampling strategy that selects the best response from a set of samples as given by a Requirement Scorer."""

    async def sample(
        self,
        action: Component,
        context: Context,
        backend: Backend,
        requirements: list[Requirement] | None,
        *,
        validation_ctx: Context | None = None,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
        show_progress: bool = True,
    ) -> SamplingResult:
        """This method performs a sampling operation based on the given instruction.

        Args:
            action : The action object to be sampled.
            context: The context to be passed to the sampling strategy.
            backend: The backend used for generating samples.
            requirements: List of requirements to test against (merged with global requirements).
            validation_ctx: Optional context to use for validation. If None, validation_ctx = ctx.
            format: output format for structured outputs.
            model_options: model options to pass to the backend during generation / validation.
            tool_calls: True if tool calls should be used during this sampling strategy.
            show_progress: if true, a tqdm progress bar is used. Otherwise, messages will still be sent to flog.

        Returns:
            SamplingResult: A result object indicating the success or failure of the sampling process.

        Raises:
            AssertionError: Asserts that all required components (repair, select_from_failure, validate, and generate) are provided before proceeding with the sampling.
        """
        validation_ctx = validation_ctx if validation_ctx is not None else context
        assert validation_ctx is not None, "Validation context must be provided."

        flog = FancyLogger.get_logger()

        sampled_results: list[ModelOutputThunk] = []
        sampled_scores: list[list[tuple[Requirement, ValidationResult]]] = []
        sampled_actions: list[Component] = []
        sample_contexts: list[Context] = []

        successful_sampled_results: list[ModelOutputThunk] = []
        successful_sampled_scores: list[list[tuple[Requirement, ValidationResult]]] = []
        successful_sampled_actions: list[Component] = []
        successful_sample_contexts: list[Context] = []

        # The `logging_redirect_tqdm` approach did not work, so instead we will use the show_progress
        # flag to determine whether we should show the pbar.
        show_progress = show_progress and flog.getEffectiveLevel() <= FancyLogger.INFO

        reqs = []
        if self.requirements is not None:
            reqs += self.requirements
        elif requirements is not None:
            reqs += requirements

        reqs = list(set(reqs))

        # check that there is exactly one ScorerRequirement
        scorer_requirements = 0
        for req in reqs:
            # strict typecheck for scorer requirement
            if isinstance(req, ScorerRequirement):
                scorer_requirements += 1

        assert scorer_requirements == 1, (
            "BestOfNSamplingStrategy requires exactly one ScorerRequirement"
        )

        loop_count = 0
        generate_loop_budget_iterator = (
            tqdm.tqdm(range(self.loop_budget))  # type: ignore
            if show_progress
            else range(self.loop_budget)  # type: ignore
        )
        validate_loop_budget_iterator = (
            tqdm.tqdm(range(self.loop_budget))  # type: ignore
            if show_progress
            else range(self.loop_budget)  # type: ignore
        )

        next_action = deepcopy(action)
        next_context = context
        flog.info("BestofNSampling Generating Loop:")
        for _ in generate_loop_budget_iterator:  # type: ignore
            loop_count += 1
            if not show_progress:
                flog.info(f"Running loop {loop_count} of {self.loop_budget}")

            # run a generation pass
            result, result_ctx = backend.generate_from_context(
                next_action,
                ctx=next_context,
                format=format,
                model_options=model_options,
                tool_calls=tool_calls,
            )
            sampled_results.append(result)
            sampled_actions.append(next_action)
            sample_contexts.append(result_ctx)

        await wait_for_all_mots(sampled_results)

        flog.info("BestofNSampling Validation Loop:")
        for i in validate_loop_budget_iterator:
            result_ctx = sample_contexts[i]
            result = sampled_results[i]
            next_action = sampled_actions[i]

            val_scores_co = mfuncs._validate(
                reqs=reqs,
                context=result_ctx,
                backend=backend,
                output=result,
                format=format,
                model_options=model_options,
                input=next_action._description,  # type: ignore
                # tool_calls=tool_calls  # Don't support using tool calls in validation strategies.
            )
            val_scores = await val_scores_co

            # match up reqs with scores
            constraint_scores = list(zip(reqs, val_scores))

            # collect all data
            sampled_scores.append(constraint_scores)

            # check if requirements pass else repair and re-sample
            # if all vals are true, save it and continue to get next sample
            if all(bool(s[1]) for s in constraint_scores):
                flog.info("SUCCESS")
                assert (
                    result._generate_log is not None
                )  # Cannot be None after generation.
                result._generate_log.is_final_result = True

                successful_sampled_results.append(result)
                successful_sampled_scores.append(constraint_scores)
                successful_sampled_actions.append(next_action)
                successful_sample_contexts.append(result_ctx)

            else:
                # log partial success and continue
                count_valid = len([s for s in constraint_scores if bool(s[1])])
                flog.info(f"FAILED. Valid: {count_valid}/{len(constraint_scores)}")

                # If we did not pass all constraints, update the instruction and try again.
                next_action, next_context = self.repair(
                    next_context,
                    result_ctx,
                    sampled_actions,
                    sampled_results,
                    sampled_scores,
                )

        # find max reward amongst results for which all requirements have passed
        if len(successful_sampled_scores) > 0:
            scores: list[float] = []
            scorer_preference_ordering = None

            for sample in successful_sampled_scores:
                for req, val_score in sample:
                    if isinstance(req, ScorerRequirement):
                        assert val_score._score is not None
                        scores.append(val_score._score)
                        scorer_preference_ordering = req.preference_ordering

            assert len(successful_sampled_results) == len(scores)
            assert scorer_preference_ordering is not None

            if scorer_preference_ordering == "max":
                best_result, best_score, best_context = max(
                    zip(successful_sampled_results, scores, successful_sample_contexts),
                    key=lambda x: x[1],
                )
            elif scorer_preference_ordering == "min":
                best_result, best_score, best_context = min(
                    zip(successful_sampled_results, scores, successful_sample_contexts),
                    key=lambda x: x[1],
                )
            else:
                raise NotImplementedError

            best_index = sampled_results.index(best_result)

            return SamplingResult(
                result_index=best_index,
                success=True,
                sample_generations=sampled_results,
                sample_validations=sampled_scores,
                sample_actions=sampled_actions,
                sample_contexts=sample_contexts,
            )

        # if all failures, call select from failure
        else:
            flog.info(
                f"Invoking select_from_failure after {len(sampled_results)} failed attempts."
            )

            # if no valid result could be determined, find a last resort.
            best_failed_index = self.select_from_failure(
                sampled_actions, sampled_results, sampled_scores
            )
            assert best_failed_index < len(sampled_results), (
                "The select_from_failure method did not return a valid result. It has to selected from failed_results."
            )
            return SamplingResult(
                result_index=best_failed_index,
                success=False,
                sample_generations=sampled_results,
                sample_validations=sampled_scores,
                sample_actions=sampled_actions,
                sample_contexts=sample_contexts,
            )

    @staticmethod
    def select_from_failure(
        sampled_actions: list[Component],
        sampled_results: list[ModelOutputThunk],
        sampled_val: list[list[tuple[Requirement, ValidationResult]]],
    ) -> int:
        """Selects the attempt with the highest score.

        Args:
            sampled_actions: List of actions that have been executed (without success).
            sampled_results: List of (unsuccessful) generation results for these actions.
            sampled_val: List of validation results for the results.

        Returns:
            The index of the result that should be selected as `.value`.
        """
        scores: list[float | None] = []

        for sample in sampled_val:
            for req, val_score in sample:
                if isinstance(req, ScorerRequirement):
                    assert val_score._score is not None
                    scores.append(val_score._score)

        assert len(sampled_results) == len(scores)

        return scores.index(max(scores))  # type: ignore

    @staticmethod
    def repair(
        old_ctx: Context,
        new_ctx: Context,
        past_actions: list[Component],
        past_results: list[ModelOutputThunk],
        past_val: list[list[tuple[Requirement, ValidationResult]]],
    ) -> tuple[Component, Context]:
        """Adds a description of the requirements that failed to a copy of the original instruction.

        Args:
            old_ctx: The context WITHOUT the last action + output.
            new_ctx: The context including the last action + output.
            past_actions: List of actions that have been executed (without success).
            past_results: List of (unsuccessful) generation results for these actions.
            past_val: List of validation results for the results.

        Returns:
            The next action component and context to be used for the next generation attempt.
        """
        pa = past_actions[-1]
        if isinstance(pa, Instruction):
            last_failed_reqs: list[Requirement] = [
                s[0] for s in past_val[-1] if not s[1]
            ]
            last_failed_reqs_str = "* " + "\n* ".join(
                [str(r.description) for r in last_failed_reqs]
            )
            return pa.copy_and_repair(
                repair_string=f"The following requirements failed before:\n{last_failed_reqs_str}"
            ), old_ctx
        return past_actions[-1], old_ctx



"""sampling methods go here."""

from .base import (
    BaseSamplingStrategy,
    MultiTurnStrategy,
    RejectionSamplingStrategy,
    RepairTemplateStrategy,
)
from .types import SamplingResult, SamplingStrategy



"""A file for helper functions that deal with OpenAI API compatible helpers."""

import json
from collections.abc import Callable
from typing import Any

from mellea.helpers.fancy_logger import FancyLogger
from mellea.stdlib.base import ModelToolCall


def extract_model_tool_requests(
    tools: dict[str, Callable], response: dict[str, Any]
) -> dict[str, ModelToolCall] | None:
    """Extracts tool calls from the dict representation of an OpenAI-like chat response object."""
    model_tool_calls: dict[str, ModelToolCall] = {}
    calls = response["message"].get("tool_calls", None)
    if calls:
        for tool_call in calls:
            tool_name = tool_call["function"]["name"]  # type: ignore
            tool_args = tool_call["function"]["arguments"]  # type: ignore

            func = tools.get(tool_name)
            if func is None:
                FancyLogger.get_logger().warning(
                    f"model attempted to call a non-existing function: {tool_name}"
                )
                continue  # skip this function if we can't find it.

            args = {}
            if tool_args is not None:
                # Returns the args as a string. Parse it here.
                args = json.loads(tool_args)
            model_tool_calls[tool_name] = ModelToolCall(tool_name, func, args)

    if len(model_tool_calls) > 0:
        return model_tool_calls
    return None


def chat_completion_delta_merge(
    chunks: list[dict], force_all_tool_calls_separate: bool = False
) -> dict:
    """Takes a list of deltas from `ChatCompletionChunk`s and merges them into a single dict representing the `ChatCompletion` choice.

    Args:
        chunks: the list of dicts that represent the message deltas
        force_all_tool_calls_separate: if `True`, tool calls in separate message deltas will not be merged (even if their index values are the same); use when providers do not return the correct index value for tool calls. If using this option, all tool calls must be fully populated in a single delta since they won't be merged.
    """
    merged: dict[str, Any] = dict()

    # `delta`s map to a single choice.
    merged["finish_reason"] = None
    merged["index"] = 0  # We always do the first choice.
    merged["logprobs"] = None
    merged["stop_reason"] = None

    # message fields
    message: dict[str, Any] = dict()
    message["content"] = ""
    message["reasoning_content"] = ""
    message["role"] = None
    m_tool_calls: list[dict] = []
    message["tool_calls"] = m_tool_calls
    merged["message"] = message

    for chunk in chunks:
        # Handle top level fields.
        if chunk.get("finish_reason", None) is not None:
            merged["finish_reason"] = chunk["finish_reason"]
        if chunk.get("stop_reason", None) is not None:
            merged["stop_reason"] = chunk["stop_reason"]

        # Handle fields of the message object.
        if message["role"] is None and chunk["delta"].get("role", None) is not None:
            message["role"] = chunk["delta"]["role"]

        if chunk["delta"].get("content", None) is not None:
            message["content"] += chunk["delta"]["content"]

        thinking = chunk["delta"].get("reasoning_content", None)
        if thinking is not None:
            message["reasoning_content"] += thinking

        tool_calls = chunk["delta"].get("tool_calls", None)
        if tool_calls is not None:
            # Merge the pieces of each tool call from separate chunks into one dict.
            # Example:
            #  chunks: [{'arguments': None, 'name': 'get_weather_precise'}, {'arguments': '{"location": "', 'name': None}, {'arguments': 'Dallas}', 'name': None}]
            #  -> [{'arguments': '{"location": "Dallas"}', 'name': 'get_weather_precise'}]
            for tool_call in tool_calls:
                idx: int = tool_call["index"]
                current_tool = None

                # In a few special cases, we want to force all tool calls to be separate regardless of the index value.
                # If not forced, check that the tool call index in the response isn't already in our list.
                create_new_tool_call = force_all_tool_calls_separate or (
                    idx > len(m_tool_calls) - 1
                )
                if create_new_tool_call:
                    current_tool = {"function": {"name": "", "arguments": None}}
                    m_tool_calls.append(current_tool)
                else:
                    # This tool has already started to be defined.
                    current_tool = m_tool_calls[idx]

                # Get the info from the function chunk.
                fx_info = tool_call["function"]
                if fx_info["name"] is not None:
                    current_tool["function"]["name"] += fx_info["name"]

                if fx_info["arguments"] is not None:
                    # Only populate args if there are any to add.
                    if current_tool["function"]["arguments"] is None:
                        current_tool["function"]["arguments"] = ""
                    current_tool["function"]["arguments"] += fx_info["arguments"]

    return merged



"""Async helper functions."""

import asyncio
from collections.abc import AsyncIterator, Coroutine
from typing import Any

from mellea.stdlib.base import ModelOutputThunk


async def send_to_queue(
    co: Coroutine[Any, Any, AsyncIterator | Any] | AsyncIterator, aqueue: asyncio.Queue
) -> None:
    """Processes the output of an async chat request by sending the output to an async queue."""
    try:
        if isinstance(co, Coroutine):
            aresponse = await co
        else:
            # Some backends (hf) don't actually return their iterator from an
            # async function. As a result, there's no coroutine to wait for here.
            aresponse = co

        if isinstance(aresponse, AsyncIterator):
            async for item in aresponse:
                await aqueue.put(item)
        else:
            await aqueue.put(aresponse)

        # Always add a sentinel value to indicate end of stream.
        await aqueue.put(None)

    # Typically, nothing awaits this function directly (only through the queue).
    # As a result, we have to be careful about catching all errors and propagating
    # them to the queue.
    except Exception as e:
        await aqueue.put(e)


async def wait_for_all_mots(mots: list[ModelOutputThunk]):
    """Helper function to make waiting for multiple ModelOutputThunks to be computed easier.

    All ModelOutputThunks must be from the same event loop. This should always be the case in sampling
    functions, session functions, and top-level mellea functions.
    """
    coroutines: list[Coroutine[Any, Any, str]] = []
    for mot in mots:
        coroutines.append(mot.avalue())

    await asyncio.gather(*coroutines)



"""OpenAI implementations for IBM's "starter pack" of Activated LoRAs."""

import asyncio
import functools
from collections.abc import Coroutine
from typing import Any

import openai
from openai.types.completion import Completion

from mellea.backends.aloras import Alora
from mellea.backends.openai import OpenAIAlora, OpenAIBackend
from mellea.backends.types import ModelOption
from mellea.helpers.async_helpers import send_to_queue
from mellea.helpers.fancy_logger import FancyLogger
from mellea.stdlib.base import GenerateType, ModelOutputThunk


class OpenAIConstraintAlora(OpenAIAlora):
    """The [Requirement Checking ALora for Granite 3.2 8B](https://huggingface.co/ibm-granite/granite-3.2-8b-alora-requirement-check) checks if the specified requirement was satisfied by the most recent model generation. Only one requirement is checked at a time."""

    def __init__(
        self, name: str, path: str, generation_prompt: str, backend: OpenAIBackend
    ):
        """Initialize after checking that the backend is correct."""
        assert backend._hf_model_id == "ibm-granite/granite-3.2-8b-instruct"
        super().__init__(name, path, generation_prompt, backend)
        # We do a lot of logging for ALoras because this is an experimental feature. Maybe we should tag these log messages?
        self._logger = FancyLogger.get_logger()

    def generate_using_strings(
        self,
        input: str,
        response: str,
        constraint: str,
        force_yn: bool = True,
        stream: bool = False,
    ) -> ModelOutputThunk:
        """Generates a constraint response from the ALora. Must be run in a running event loop."""
        # Go ahead and do runtime type-checking because passing CBlocks into this function is a common error.
        assert type(input) is str
        assert type(response) is str
        assert type(constraint) is str

        # Params aren't needed when just getting the backend args.
        backend_model_opts = self._backend._simplify_and_merge(None, False)
        sys_prompt = backend_model_opts.get(ModelOption.SYSTEM_PROMPT, None)

        chat = [
            *([{"role": "system", "content": sys_prompt}] if sys_prompt else []),
            {"role": "user", "content": input},
            {"role": "assistant", "content": response},
        ]

        prompt = self._backend.apply_chat_template(chat)
        prompt += f"\nRequirement: {constraint}<|end_of_text|>\n"  # type: ignore
        prompt += self._generation_prompt

        self._logger.debug(f"Prompt for non-cached aLoRA({self.name}):\n{prompt}")

        force_yn_args = {}
        if force_yn:
            assert hasattr(self._backend, "_tokenizer")
            token_Y = self._backend._tokenizer("Y", add_special_tokens=False)[
                "input_ids"
            ][0]  # type: ignore
            token_N = self._backend._tokenizer("N", add_special_tokens=False)[
                "input_ids"
            ][0]  # type: ignore

            force_yn_args["logit_bias"] = {str(token_Y): 100, str(token_N): 100}

        chat_response: Coroutine[
            Any, Any, openai.AsyncStream[Completion] | Completion
        ] = self._backend._async_client.completions.create(
            model=self.name,
            prompt=prompt,
            max_tokens=1,
            n=1,
            stream=stream,
            **force_yn_args,
        )  # type: ignore

        output = ModelOutputThunk(None)
        output._meta["alora_name"] = self.name

        output._process = processing
        output._post_process = functools.partial(post_processing, backend=self._backend)

        try:
            # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
            # We can also support synchronous calls by adding a flag and changing this ._generate function.

            # This function should always be called from a running event loop so we don't have to worry about
            # scheduling the task to a specific event loop here.
            output._generate = asyncio.create_task(
                send_to_queue(chat_response, output._async_queue)
            )
            output._generate_type = GenerateType.ASYNC
        except RuntimeError as e:
            # Most likely cause is running this function without an event loop present
            raise e

        return output


async def processing(mot: ModelOutputThunk, chunk: Completion):
    """Called to process the incoming chunks."""
    if mot._underlying_value is None:
        mot._underlying_value = ""
    mot._underlying_value += chunk.choices[0].text


async def post_processing(backend: OpenAIBackend, mot: ModelOutputThunk):
    """Called after all data has been received."""
    backend.formatter.parse(mot._action, mot)  # type: ignore


def add_granite_aloras(backend: OpenAIBackend):
    """Adds the IBM Granite "starter pack" ALoras to a backend."""
    backend.add_alora(
        OpenAIConstraintAlora(
            name="constraint",
            path="ibm-granite/granite-3.2-8b-alora-requirement-check",
            generation_prompt="<|start_of_role|>check_requirement<|end_of_role|>",
            backend=backend,
        )
    )
